from keras.preprocessing.image import load_img, img_to_array
import numpy as np
import scipy
from keras import backend as K

from io import BytesIO
import PIL.Image
# from IPython.display import clear_output, Image, display

# We are in the "test" phase, not the "training" phase, w.r.t. to the model we are analyzing
K.set_learning_phase(0)

# from keras.applications import inception_v3
# Load InceptionV3 with ImageNet pre-trained weights
# model = inception_v3.InceptionV3(weights='imagenet', include_top=False)

import inception_v1
# Load InceptionV1 with ImageNet pre-trained weights
# Adapted from: https://github.com/fchollet/deep-learning-models/pull/59
model = inception_v1.InceptionV1(weights='imagenet', include_top=False)

model.summary()

settings_InceptionV3 = {
    'features': {
        'mixed2': 2.0,
        'mixed3': 0.5,
        'mixed4': 1.0,
        'mixed5': 2.0,
    },
}

settings_InceptionV1 = {
    'features': {
        'Mixed_4c_Concatenated': 1.0,
    },
}

settings = settings_InceptionV1

index = 0

def preprocess_image(image_path):
    img = load_img(image_path)
    img = img_to_array(img)
    img = np.expand_dims(img, axis=0) # Inserts new axis at position 0 - TO DO: check img.shape
    # img = inception_v3.preprocess_input(img) # 3D -> 1D
    img = inception_v1.preprocess_input(img) # 3D -> 1D
    return img

def deprocess_image(x):
    # Check ~/.keras/keras.json to make sure "image_data_format" is set to "channels_last"
    # or print(K.image_data_format())
    x = x.reshape((x.shape[1], x.shape[2], 3)) # 1D -> 3D
    x /= 2. # Normalize from -1/1 to 0/255
    x += 0.5 # Normalize from -1/1 to 0/255
    x *= 255. # Normalize from -1/1 to 0/255
    x = np.clip(x, 0, 255).astype('uint8') # Clip to visible range
    return x

# Simple resize function based on scipy
def resize_img(img, size):
    img = np.copy(img)
    factors = (1, float(size[0]) / img.shape[1], float(size[1]) / img.shape[2], 1)
    return scipy.ndimage.zoom(img, factors, order=1)

# Simple save function based on scipy
def save_image(img, fname):
    pil_img = deprocess_image(np.copy(img))
    scipy.misc.imsave(fname, pil_img)

def show_image(img, fmt='jpeg'):
    img = deprocess_image(np.copy(img))
    f = BytesIO()
    PIL.Image.fromarray(img).save(f, fmt)
    display(Image(data=f.getvalue()))

def deep_dream():

    global index

    # Dictionary of the names of the layers and the layers themselves
    layer_dict = dict([(layer.name, layer) for layer in model.layers])

    # Input of the first layer, in the example: model.input
    dream = model.layers[0].input

    # A TF variable (persistent)
    loss = K.variable(0.)

    # Iterate only over the layers picked above
    for layer_name in settings['features']:
        print(layer_name)
        assert layer_name in layer_dict.keys(), 'Layer ' + layer_name + ' not found in model.' # Layer in model?
        coeff = settings['features'][layer_name] # Coefficient for the layer
        x = layer_dict[layer_name].output # Output of the current layer
        # We avoid border artifacts by only involving non-border pixels in the loss, offset by 2 on all sides
        scaling = K.prod(K.cast(K.shape(x), 'float32'))
        loss += coeff * K.sum(K.square(x[:, 2: -2, 2: -2, :])) / scaling # L2 norm

    # Compute the gradients of the dream w.r.t. the loss.
    grads = K.gradients(loss, dream)[0]
    # Normalize gradients.
    grads /= K.maximum(K.mean(K.abs(grads)), K.epsilon())

    fetch_loss_and_grads = K.function([dream], [loss, grads])
    def gradient_ascent(x, iterations, step, max_loss=None):
        for i in range(iterations):
            outs = fetch_loss_and_grads([x])
            loss_value = outs[0]
            grad_values = outs[1]
            if max_loss is not None and loss_value > max_loss:
                break
            # print('..Loss value at', i, ':', loss_value)
            x += step * grad_values
        return x

    step = 0.01  # Gradient ascent step size
    num_octave = 10  # Number of scales at which to run gradient ascent
    octave_scale = 1.2  # Size ratio between scales
    iterations = 10  # Number of ascent steps per scale
    max_loss = 10.

    img = preprocess_image('/Users/adamjahnke/Desktop/MAT_185_MASTER/test_heads/image' + str(index) +'.jpg') #/Volumes/pi/Face_Sound/images/img.jpg #3-deepdream/input.jpg
    print(img.shape)
    original_shape = img.shape[1:3]
    print(original_shape)
    successive_shapes = [original_shape]

    for i in range(1, num_octave):
        shape = tuple([int(dim / (octave_scale ** i)) for dim in original_shape])
        print(shape)
        successive_shapes.append(shape)
    successive_shapes = successive_shapes[::-1] # Reverse the list so we go small to original

    original_img = np.copy(img)
    shrunk_original_img = resize_img(img, successive_shapes[0])

    for shape in successive_shapes:
        print('Processing image shape', shape)

        # 1. Resize current image to current shape
        img = resize_img(img, shape)
        # 2. Optimize current image w.r.t. selected layers
        img = gradient_ascent(img, iterations=iterations, step=step, max_loss=max_loss)
        # 3. Resize ("upscale") previous shape original image to current shape
        upscaled_shrunk_original_img = resize_img(shrunk_original_img, shape)
        # 4. Resize ("downscale") original shape original image to current shape
        same_size_original = resize_img(original_img, shape)
        # 5. Find the details that are lost in the upscaling process
        lost_detail = same_size_original - upscaled_shrunk_original_img
        # 6. Add these details back to the optimized image
        img += lost_detail # Add back the lost details to the optimized image

        #

        # Prepare for next iterationy
        shrunk_original_img = resize_img(original_img, shape) #

    # Save final image
    save_image(img, fname='/Users/adamjahnke/Desktop/MAT_185_MASTER/deepdream_output/output' + str(index) +'.png')
    # show_image(img)

    index += 1

while True:
#for i in range(30):
    deep_dream()
